@@ -28,6 +28,7 @@
 #include "networkd.h"
 #include "parse.h"
 #include "util.h"
+#include "validation.h"
 
 /**
  * Append WiFi frequencies to wpa_supplicant's freq_list=
@@ -151,6 +152,69 @@ write_tunnel_params(GString* s, const NetplanNetDefinition* def)
     g_string_free(params, TRUE);
 }
 
+static void
+write_wireguard_params(GString* s, const NetplanNetDefinition* def)
+{
+    GString *params = NULL;
+    params = g_string_sized_new(200);
+
+    g_assert(def->tunnel.private_key);
+    /* The "PrivateKeyFile=" setting is available as of systemd-netwokrd v242+
+     * Base64 encoded PrivateKey= or absolute PrivateKeyFile= fields are mandatory.
+     *
+     * The key was already validated via validate_tunnel_grammar(), but we need
+     * to differentiate between base64 key VS absolute path key-file. And a base64
+     * string could (theoretically) start with '/', so we use is_wireguard_key()
+     * as well to check for more specific characteristics (if needed). */
+    if (def->tunnel.private_key[0] == '/' && !is_wireguard_key(def->tunnel.private_key))
+        g_string_append_printf(params, "PrivateKeyFile=%s\n", def->tunnel.private_key);
+    else
+        g_string_append_printf(params, "PrivateKey=%s\n", def->tunnel.private_key);
+
+    if (def->tunnel.port)
+        g_string_append_printf(params, "ListenPort=%u\n", def->tunnel.port);
+    /* This is called FirewallMark= as of systemd v243, but we keep calling it FwMark= for
+       backwards compatibility. FwMark= is still supported, but deprecated:
+       https://github.com/systemd/systemd/pull/12478 */
+    if (def->tunnel.fwmark)
+        g_string_append_printf(params, "FwMark=%u\n", def->tunnel.fwmark);
+
+    g_string_append_printf(s, "\n[WireGuard]\n%s", params->str);
+    g_string_free(params, TRUE);
+
+    for (guint i = 0; i < def->wireguard_peers->len; i++) {
+        NetplanWireguardPeer *peer = g_array_index (def->wireguard_peers, NetplanWireguardPeer*, i);
+        GString *peer_s = g_string_sized_new(200);
+
+        g_string_append_printf(peer_s, "PublicKey=%s\n", peer->public_key);
+        g_string_append(peer_s, "AllowedIPs=");
+        for (guint i = 0; i < peer->allowed_ips->len; ++i) {
+            if (i > 0 )
+                g_string_append_c(peer_s, ',');
+            g_string_append_printf(peer_s, "%s", g_array_index(peer->allowed_ips, char*, i));
+        }
+        g_string_append_c(peer_s, '\n');
+
+        if (peer->keepalive)
+            g_string_append_printf(peer_s, "PersistentKeepalive=%d\n", peer->keepalive);
+        if (peer->endpoint)
+            g_string_append_printf(peer_s, "Endpoint=%s\n", peer->endpoint);
+        /* The key was already validated via validate_tunnel_grammar(), but we need
+         * to differentiate between base64 key VS absolute path key-file. And a base64
+         * string could (theoretically) start with '/', so we use is_wireguard_key()
+         * as well to check for more specific characteristics (if needed). */
+        if (peer->preshared_key) {
+            if (peer->preshared_key[0] == '/' && !is_wireguard_key(peer->preshared_key))
+                g_string_append_printf(peer_s, "PresharedKeyFile=%s\n", peer->preshared_key);
+            else
+                g_string_append_printf(peer_s, "PresharedKey=%s\n", peer->preshared_key);
+        }
+
+        g_string_append_printf(s, "\n[WireGuardPeer]\n%s", peer_s->str);
+        g_string_free(peer_s, TRUE);
+    }
+}
+
 static void
 write_link_file(const NetplanNetDefinition* def, const char* rootdir, const char* path)
 {
@@ -326,9 +390,9 @@ write_netdev_file(const NetplanNetDefinition* def, const char* rootdir, const ch
                 case NETPLAN_TUNNEL_MODE_SIT:
                 case NETPLAN_TUNNEL_MODE_VTI:
                 case NETPLAN_TUNNEL_MODE_VTI6:
-                    g_string_append_printf(s,
-                                          "Kind=%s\n",
-                                          tunnel_mode_to_string(def->tunnel.mode));
+                case NETPLAN_TUNNEL_MODE_WIREGUARD:
+                    g_string_append_printf(s, "Kind=%s\n",
+                                           tunnel_mode_to_string(def->tunnel.mode));
                     break;
 
                 case NETPLAN_TUNNEL_MODE_IP6IP6:
@@ -341,14 +405,13 @@ write_netdev_file(const NetplanNetDefinition* def, const char* rootdir, const ch
                     g_assert_not_reached();
                 // LCOV_EXCL_STOP
             }
-
-            write_tunnel_params(s, def);
+            if (def->tunnel.mode == NETPLAN_TUNNEL_MODE_WIREGUARD)
+                write_wireguard_params(s, def);
+            else
+                write_tunnel_params(s, def);
             break;
 
-        // LCOV_EXCL_START
-        default:
-            g_assert_not_reached();
-        // LCOV_EXCL_STOP
+        default: g_assert_not_reached(); // LCOV_EXCL_LINE
     }
 
     /* these do not contain secrets and need to be readable by
@@ -402,6 +465,19 @@ write_ip_rule(NetplanIPRule* r, GString* s)
         g_string_append_printf(s, "TypeOfService=%d\n", r->tos);
 }
 
+static void
+write_addr_option(NetplanAddressOptions* o, GString* s)
+{
+    g_string_append_printf(s, "\n[Address]\n");
+    g_assert(o->address);
+    g_string_append_printf(s, "Address=%s\n", o->address);
+
+    if (o->lifetime)
+        g_string_append_printf(s, "PreferredLifetime=%s\n", o->lifetime);
+    if (o->label)
+        g_string_append_printf(s, "Label=%s\n", o->label);
+}
+
 #define DHCP_OVERRIDES_ERROR                                            \
     "ERROR: %s: networkd requires that %s has the same value in both "  \
     "dhcp4_overrides and dhcp6_overrides\n"
@@ -459,7 +535,10 @@ combine_dhcp_overrides(const NetplanNetDefinition* def, NetplanDHCPOverrides* co
     }
 }
 
-static void
+/**
+ * Write the needed networkd .network configuration for the selected netplan definition.
+ */
+void
 write_network_file(const NetplanNetDefinition* def, const char* rootdir, const char* path)
 {
     GString* network = NULL;
@@ -524,13 +603,13 @@ write_network_file(const NetplanNetDefinition* def, const char* rootdir, const c
     if (def->ip6_addresses)
         for (unsigned i = 0; i < def->ip6_addresses->len; ++i)
             g_string_append_printf(network, "Address=%s\n", g_array_index(def->ip6_addresses, char*, i));
-    if (def->ip6_addr_gen_mode) {
-        /* TODO: Figure out how we can configure ipv6-address-generation for networkd.
-         *       IPv6Token= seems to be the corresponding option, but it doesn't do
-         *       exactly what we need and has quite some restrictions, c.f.:
-         *       https://github.com/systemd/systemd/issues/4625
-         *       https://github.com/systemd/systemd/pull/14415 */
-        g_fprintf(stderr, "ERROR: %s: ipv6-address-generation is not supported by networkd\n", def->id);
+    if (def->ip6_addr_gen_token) {
+        g_string_append_printf(network, "IPv6Token=static:%s\n", def->ip6_addr_gen_token);
+    } else if (def->ip6_addr_gen_mode > NETPLAN_ADDRGEN_EUI64) {
+        /* EUI-64 mode is enabled by default, if no IPv6Token= is specified */
+        /* TODO: Enable stable-privacy mode for networkd, once PR#16618 has been released:
+         *       https://github.com/systemd/systemd/pull/16618 */
+        g_fprintf(stderr, "ERROR: %s: ipv6-address-generation mode is not supported by networkd\n", def->id);
         exit(1);
     }
     if (def->accept_ra == NETPLAN_RA_MODE_ENABLED)
@@ -563,7 +642,7 @@ write_network_file(const NetplanNetDefinition* def, const char* rootdir, const c
     if (def->type >= NETPLAN_DEF_TYPE_VIRTUAL)
         g_string_append(network, "ConfigureWithoutCarrier=yes\n");
 
-    if (def->bridge) {
+    if (def->bridge && def->backend != NETPLAN_BACKEND_OVS) {
         g_string_append_printf(network, "Bridge=%s\n", def->bridge);
 
         if (def->bridge_params.path_cost || def->bridge_params.port_priority)
@@ -573,14 +652,14 @@ write_network_file(const NetplanNetDefinition* def, const char* rootdir, const c
         if (def->bridge_params.port_priority)
             g_string_append_printf(network, "Priority=%u\n", def->bridge_params.port_priority);
     }
-    if (def->bond) {
+    if (def->bond && def->backend != NETPLAN_BACKEND_OVS) {
         g_string_append_printf(network, "Bond=%s\n", def->bond);
 
         if (def->bond_params.primary_slave)
             g_string_append_printf(network, "PrimarySlave=true\n");
     }
 
-    if (def->has_vlans) {
+    if (def->has_vlans && def->backend != NETPLAN_BACKEND_OVS) {
         /* iterate over all netdefs to find VLANs attached to us */
         GList *l = netdefs_ordered;
         const NetplanNetDefinition* nd;
@@ -604,6 +683,13 @@ write_network_file(const NetplanNetDefinition* def, const char* rootdir, const c
         }
     }
 
+    if (def->address_options) {
+        for (unsigned i = 0; i < def->address_options->len; ++i) {
+            NetplanAddressOptions* opts = g_array_index(def->address_options, NetplanAddressOptions*, i);
+            write_addr_option(opts, network);
+        }
+    }
+
     if (def->dhcp4 || def->dhcp6 || def->critical) {
         /* NetworkManager compatible route metrics */
         g_string_append(network, "\n[DHCP]\n");
@@ -812,19 +898,8 @@ write_wpa_unit(const NetplanNetDefinition* def, const char* rootdir)
 {
     g_autoptr(GError) err = NULL;
     g_autofree gchar *stdouth = NULL;
-    g_autofree gchar *stderrh = NULL;
-    gint exit_status = 0;
 
-    gchar *argv[] = {"bin" "/" "systemd-escape", def->id, NULL};
-    g_spawn_sync("/", argv, NULL, 0, NULL, NULL, &stdouth, &stderrh, &exit_status, &err);
-    g_spawn_check_exit_status(exit_status, &err);
-    if (err != NULL) {
-        // LCOV_EXCL_START
-        g_fprintf(stderr, "failed to ask systemd to escape %s; exit %d\nstdout: '%s'\nstderr: '%s'", def->id, exit_status, stdouth, stderrh);
-        exit(1);
-        // LCOV_EXCL_STOP
-    }
-    g_strstrip(stdouth);
+    stdouth = systemd_escape(def->id);
 
     GString* s = g_string_new("[Unit]\n");
     g_autofree char* path = g_strjoin(NULL, "/run/systemd/system/netplan-wpa-", stdouth, ".service", NULL);
@@ -863,6 +938,9 @@ write_wpa_conf(const NetplanNetDefinition* def, const char* rootdir)
             if (ap->bssid) {
                 g_string_append_printf(s, "  bssid=%s\n", ap->bssid);
             }
+            if (ap->hidden) {
+                g_string_append(s, "  scan_ssid=1\n");
+            }
             if (ap->band == NETPLAN_WIFI_BAND_24) {
                 // initialize 2.4GHz frequency hashtable
                 if(!wifi_frequency_24)
@@ -990,8 +1068,12 @@ cleanup_networkd_conf(const char* rootdir)
 {
     unlink_glob(rootdir, "/run/systemd/network/10-netplan-*");
     unlink_glob(rootdir, "/run/netplan/wpa-*.conf");
+    unlink_glob(rootdir, "/run/systemd/system/systemd-networkd.service.wants/netplan-wpa-*.service");
     unlink_glob(rootdir, "/run/systemd/system/netplan-wpa-*.service");
     unlink_glob(rootdir, "/run/udev/rules.d/99-netplan-*");
+    /* Historically (up to v0.98) we had netplan-wpa@*.service files, in case of an
+     * upgraded system, we need to make sure to clean those up. */
+    unlink_glob(rootdir, "/run/systemd/system/systemd-networkd.service.wants/netplan-wpa@*.service");
 }
 
 /**
