@@ -28,6 +28,7 @@
 #include "nm.h"
 #include "parse.h"
 #include "util.h"
+#include "validation.h"
 
 GString* udev_rules;
 
@@ -111,6 +112,8 @@ type_str(const NetplanNetDefinition* def)
         case NETPLAN_DEF_TYPE_VLAN:
             return "vlan";
         case NETPLAN_DEF_TYPE_TUNNEL:
+            if (def->tunnel.mode == NETPLAN_TUNNEL_MODE_WIREGUARD)
+                return "wireguard";
             return "ip-tunnel";
         // LCOV_EXCL_START
         default:
@@ -201,23 +204,14 @@ write_routes(const NetplanNetDefinition* def, GString *s, int family)
                 exit(1);
             }
 
-            if (cur_route->scope && g_ascii_strcasecmp(cur_route->scope, "global") != 0) {
-                g_fprintf(stderr, "ERROR: %s: NetworkManager only supports global scoped routes\n", def->id);
-                exit(1);
-            }
-
-            if (cur_route->table != NETPLAN_ROUTE_TABLE_UNSPEC) {
-                g_fprintf(stderr, "ERROR: %s: NetworkManager does not support non-default routing tables\n", def->id);
-                exit(1);
-            }
-
-            if (cur_route->from) {
-                g_fprintf(stderr, "ERROR: %s: NetworkManager does not support routes with 'from'\n", def->id);
-                exit(1);
-            }
-
-            if (cur_route->onlink) {
-                g_fprintf(stderr, "ERROR: %s: NetworkManager does not support on-link routes\n", def->id);
+            if (!g_strcmp0(cur_route->scope, "global")) {
+                /* For IPv6 addresses, kernel and NetworkManager don't support a scope.
+                 * For IPv4 addresses, NetworkManager determines the scope of addresses on its own
+                 * ("link" for addresses without gateway, "global" for addresses with next-hop). */
+                g_debug("%s: NetworkManager does not support setting a scope for routes, it will auto-detect them.", def->id);
+            } else if (cur_route->scope) {
+                /* Error out if scope is not set to its default value of 'global' */
+                g_fprintf(stderr, "ERROR: %s: NetworkManager does not support setting a scope for routes\n", def->id);
                 exit(1);
             }
 
@@ -226,6 +220,21 @@ write_routes(const NetplanNetDefinition* def, GString *s, int family)
             if (cur_route->metric != NETPLAN_METRIC_UNSPEC)
                 g_string_append_printf(s, ",%d", cur_route->metric);
             g_string_append(s, "\n");
+
+            if (   cur_route->onlink
+                || cur_route->table != NETPLAN_ROUTE_TABLE_UNSPEC
+                || cur_route->from) {
+                g_string_append_printf(s, "route%d_options=", j);
+                if (cur_route->onlink) {
+                    /* onlink for IPv6 addresses is only supported since nm-1.18.0. */
+                    g_string_append_printf(s, "onlink=true,");
+                }
+                if (cur_route->table != NETPLAN_ROUTE_TABLE_UNSPEC)
+                    g_string_append_printf(s, "table=%u,", cur_route->table);
+                if (cur_route->from)
+                    g_string_append_printf(s, "src=%s,", cur_route->from);
+                s->str[s->len - 1] = '\n';
+            }
             j++;
         }
     }
@@ -321,6 +330,60 @@ write_bridge_params(const NetplanNetDefinition* def, GString *s)
     }
 }
 
+static void
+write_wireguard_params(const NetplanNetDefinition* def, GString *s)
+{
+    g_assert(def->tunnel.private_key);
+    g_string_append(s, "\n[wireguard]\n");
+
+    /* The key was already validated via validate_tunnel_grammar(), but we need
+     * to differentiate between base64 key VS absolute path key-file. And a base64
+     * string could (theoretically) start with '/', so we use is_wireguard_key()
+     * as well to check for more specific characteristics (if needed). */
+    if (def->tunnel.private_key[0] == '/' && !is_wireguard_key(def->tunnel.private_key)) {
+        g_fprintf(stderr, "%s: private key needs to be base64 encoded when using the NM backend\n", def->id);
+        exit(1);
+    } else
+        g_string_append_printf(s, "private-key=%s\n", def->tunnel.private_key);
+
+    if (def->tunnel.port)
+        g_string_append_printf(s, "listen-port=%u\n", def->tunnel.port);
+    if (def->tunnel.fwmark)
+        g_string_append_printf(s, "fwmark=%u\n", def->tunnel.fwmark);
+
+    for (guint i = 0; i < def->wireguard_peers->len; i++) {
+        NetplanWireguardPeer *peer = g_array_index (def->wireguard_peers, NetplanWireguardPeer*, i);
+        g_assert(peer->public_key);
+        g_string_append_printf(s, "\n[wireguard-peer.%s]\n", peer->public_key);
+
+        if (peer->keepalive)
+            g_string_append_printf(s, "persistent-keepalive=%d\n", peer->keepalive);
+        if (peer->endpoint)
+            g_string_append_printf(s, "endpoint=%s\n", peer->endpoint);
+        /* The key was already validated via validate_tunnel_grammar(), but we need
+         * to differentiate between base64 key VS absolute path key-file. And a base64
+         * string could (theoretically) start with '/', so we use is_wireguard_key()
+         * as well to check for more specific characteristics (if needed). */
+        if (peer->preshared_key) {
+            if (peer->preshared_key[0] == '/' && !is_wireguard_key(peer->preshared_key)) {
+                g_fprintf(stderr, "%s: shared key needs to be base64 encoded when using the NM backend\n", def->id);
+                exit(1);
+            } else {
+                g_string_append_printf(s, "preshared-key=%s\n", peer->preshared_key);
+                g_string_append(s, "preshared-key-flags=0\n");
+            }
+        }
+        if (peer->allowed_ips && peer->allowed_ips->len > 0) {
+            g_string_append(s, "allowed-ips=");
+            for (guint i = 0; i < peer->allowed_ips->len; ++i) {
+                if (i > 0 ) g_string_append_c(s, ';');
+                g_string_append_printf(s, "%s", g_array_index(peer->allowed_ips, char*, i));
+            }
+            g_string_append_c(s, '\n');
+        }
+    }
+}
+
 static void
 write_tunnel_params(const NetplanNetDefinition* def, GString *s)
 {
@@ -437,6 +500,7 @@ write_nm_conf_access_point(NetplanNetDefinition* def, const char* rootdir, const
     g_autofree char* conf_path = NULL;
     mode_t orig_umask;
     char uuidstr[37];
+    const char *match_interface_name = NULL;
 
     if (def->type == NETPLAN_DEF_TYPE_WIFI)
         g_assert(ap);
@@ -472,12 +536,10 @@ write_nm_conf_access_point(NetplanNetDefinition* def, const char* rootdir, const
         else if (!def->has_match)
             g_string_append_printf(s, "interface-name=%s\n", def->id);
         else if (def->match.original_name) {
-            /* NM does not support interface name globbing */
-            if (strpbrk(def->match.original_name, "*[]?")) {
-                g_fprintf(stderr, "ERROR: %s: NetworkManager definitions do not support name globbing\n", def->id);
-                exit(1);
-            }
-            g_string_append_printf(s, "interface-name=%s\n", def->match.original_name);
+            if (strpbrk(def->match.original_name, "*[]?"))
+                match_interface_name = def->match.original_name;
+            else
+                g_string_append_printf(s, "interface-name=%s\n", def->match.original_name);
         }
         /* else matches on something other than the name, do not restrict interface-name */
     } else {
@@ -609,8 +671,17 @@ write_nm_conf_access_point(NetplanNetDefinition* def, const char* rootdir, const
     if (def->type == NETPLAN_DEF_TYPE_BOND)
         write_bond_parameters(def, s);
 
-    if (def->type == NETPLAN_DEF_TYPE_TUNNEL)
-        write_tunnel_params(def, s);
+    if (def->type == NETPLAN_DEF_TYPE_TUNNEL) {
+        if (def->tunnel.mode == NETPLAN_TUNNEL_MODE_WIREGUARD)
+            write_wireguard_params(def, s);
+        else
+            write_tunnel_params(def, s);
+    }
+
+    if (match_interface_name) {
+        g_string_append(s, "\n[match]\n");
+        g_string_append_printf(s, "interface-name=%s;\n", match_interface_name);
+    }
 
     g_string_append(s, "\n[ipv4]\n");
 
@@ -660,7 +731,11 @@ write_nm_conf_access_point(NetplanNetDefinition* def, const char* rootdir, const
         if (def->ip6_addresses)
             for (unsigned i = 0; i < def->ip6_addresses->len; ++i)
                 g_string_append_printf(s, "address%i=%s\n", i+1, g_array_index(def->ip6_addresses, char*, i));
-        if (def->ip6_addr_gen_mode) {
+        if (def->ip6_addr_gen_token) {
+            /* Token implies EUI-64, i.e mode=0 */
+            g_string_append(s, "addr-gen-mode=0\n");
+            g_string_append_printf(s, "token=%s\n", def->ip6_addr_gen_token);
+        } else if (def->ip6_addr_gen_mode) {
             g_string_append_printf(s, "addr-gen-mode=%s\n", addr_gen_mode_str(def->ip6_addr_gen_mode));
         }
         if (def->ip6_privacy)
@@ -700,6 +775,9 @@ write_nm_conf_access_point(NetplanNetDefinition* def, const char* rootdir, const
         if (ap->bssid) {
             g_string_append_printf(s, "bssid=%s\n", ap->bssid);
         }
+        if (ap->hidden) {
+            g_string_append(s, "hidden=true\n");
+        }
         if (ap->band == NETPLAN_WIFI_BAND_5 || ap->band == NETPLAN_WIFI_BAND_24) {
             g_string_append_printf(s, "band=%s\n", wifi_band_str(ap->band));
             /* Channel is only unambiguous, if band is set. */
@@ -747,6 +825,11 @@ write_nm_conf(NetplanNetDefinition* def, const char* rootdir)
         exit(1);
     }
 
+    if (def->address_options) {
+        g_fprintf(stderr, "ERROR: %s: NetworkManager does not support address options\n", def->id);
+        exit(1);
+    }
+
     /* for wifi we need to create a separate connection file for every SSID */
     if (def->type == NETPLAN_DEF_TYPE_WIFI) {
         GHashTableIter iter;
